from math import nan, sqrt
from numpy.random import RandomState
import numpy as np
from sympy import preorder_traversal

def complexity_sympy(model):
    c=0
    for _ in preorder_traversal(model):
        c += 1
    return c

# described in Apendix 4 of paper Contemporary Symbolic Regression Methods and their Relative Performance
def noisefy(y, noise_level, random_state):
    yRMSE = 0
    for i in range(len(y)):
        yRMSE+=(y[i]*y[i])
    yRMSE=sqrt(yRMSE/len(y))
    yRMSE_noise_SD = noise_level*yRMSE
    rg = RandomState(random_state)
    noise = rg.normal(0, yRMSE_noise_SD, len(y))
    y_n = []
    for i in range(len(y)):
        y_n.append(y[i]+noise[i])
    return y_n

def logistic(yp):
    return 1.0/(1.0+np.exp(-yp))

def binarize(yp):
    #yp = logistic(yp)
    return (yp > 0.5)*1

def proba(yp):
    yp = logistic(yp)
    return np.vstack([1 - yp, yp]).T


import numpy as np
from sklearn.base import BaseEstimator
from sklearn.metrics import r2_score, accuracy_score
from sympy import sympify, simplify
import rils_rols_cpp
from .utils import binarize, proba, complexity_sympy
import warnings
import multiprocessing.pool
import functools
warnings.filterwarnings("ignore")

class RILSROLSBase(BaseEstimator):

    def __init__(self, classification=None, max_fit_calls=100000, max_seconds=100, complexity_penalty=0.001, max_complexity=200, sample_size=0.1, verbose=False, random_state=0):
        self.classification = classification
        self.max_seconds = max_seconds
        self.max_fit_calls = max_fit_calls
        self.max_complexity = max_complexity
        self.complexity_penalty = complexity_penalty
        self.sample_size = sample_size
        self.verbose = verbose
        self.random_state = random_state
        self.rr_cpp = None
        self.model = None
        self.model_simp = None
    
    def timeout(max_timeout):
        """Timeout decorator, parameter in seconds."""
        def timeout_decorator(item):
            """Wrap the original function."""
            @functools.wraps(item)
            def func_wrapper(*args, **kwargs):
                """Closure for function."""
                pool = multiprocessing.pool.ThreadPool(processes=1)
                async_result = pool.apply_async(item, args, kwargs)
                # raises a TimeoutError if execution exceeds max_timeout
                return async_result.get(max_timeout)
            return func_wrapper
        return timeout_decorator

    @timeout(2.0)
    def call_model_symplify(self):
        self.model_simp = simplify(self.model, ratio=1)

    def fit(self, X, y):
        self.rr_cpp = rils_rols_cpp.rils_rols(self.classification,int(self.max_fit_calls),int(self.max_seconds),self.complexity_penalty,self.max_complexity,self.sample_size,self.verbose,int(self.random_state))
        X = np.array(X)
        data_cnt = X.shape[0]
        feat_cnt = X.shape[1]
        X.resize((data_cnt*feat_cnt, 1))
        y = np.array(y)
        # now call CPP fit method to finish the job
        self.rr_cpp.fit(X, y, data_cnt, feat_cnt)
        self.model = self.rr_cpp.get_model_string()
        self.best_time = self.rr_cpp.get_best_time()
        self.total_time = self.rr_cpp.get_total_time()
        self.fit_calls = self.rr_cpp.get_fit_calls()
        if self.verbose:
            print('Finished otimization, doing final symplification...')
        try:
            self.call_model_symplify()
        except:
            # otherwise, just sympify it -- this does not stuck
            if self.verbose:
                print("Simplification failed within given timeout, so just doing sympify.")
            self.model_simp = sympify(self.model)
        return (self.model, self.model_simp)
    
    def check_model(self):
        if self.model is None or self.rr_cpp is None:
            raise Exception("Cannot predict because model is not build yet. First call fit().")
        
    def predict(self, X):
        self.check_model()
        X = np.array(X)
        data_cnt = X.shape[0]
        feat_cnt = X.shape[1]
        X.resize((data_cnt*feat_cnt, 1))
        y = self.rr_cpp.predict(X, data_cnt, feat_cnt)
        return y

    def model_string(self):
        self.check_model()
        return self.model_simp

    def fit_report_string(self):
        self.check_model()
        return "maxTime={0}\tmaxFitCalls={1}\tseed={2}\tsizePenalty={3}\tmaxComplexity={4}\tsampleShare={5}\ttotalTime={6:.1f}\tbestTime={7}\tfitCalls={8}\tsimpSize={9}\texpr={10}\texprSimp={11}".format(
            self.max_seconds,self.max_fit_calls,self.random_state,self.complexity_penalty,self.max_complexity, self.sample_size, self.total_time,self.best_time, self.fit_calls, complexity_sympy(self.model_simp),  self.model, self.model_simp)

class RILSROLSRegressor(RILSROLSBase):
        
    def __init__(self, max_fit_calls=100000, max_seconds=100, complexity_penalty=0.001, max_complexity=200, sample_size=0.1, verbose=False, random_state=0):
        super().__init__(False, max_fit_calls, max_seconds, complexity_penalty, max_complexity, sample_size, verbose,  random_state)

    def score(self, X, y):
        yp = self.predict(X)
        return r2_score(y, yp)

class RILSROLSBinaryClassifier(RILSROLSBase):
    
    def __init__(self, max_fit_calls=100000, max_seconds=100, complexity_penalty=0.001, max_complexity=200, sample_size=0.1, verbose=False, random_state=0):
        super().__init__(True, max_fit_calls, max_seconds, complexity_penalty, max_complexity, sample_size, verbose,  random_state)

    def check_binary_targets(self, y):
        for yi in y:
            if yi!=0 and yi!=1:
                raise Exception('The classifier works only for binary targets, so allowed target values are 0 or 1.')

    def predict(self, X):
        preds = super().predict(X)
        return binarize(preds)
    
    def predict_proba(self, X):
        preds = self.predict(X)
        return proba(preds)
    
    def score(self, X, y):
        self.check_binary_targets(y)
        yp = self.predict(X)
        return accuracy_score(y, yp)
    
    def fit(self, X, y):
        self.check_binary_targets(y)
        return super().fit(X, y)
    


from setuptools import setup, find_packages
from pybind11.setup_helpers import Pybind11Extension, build_ext
import os

#with open("LICENSE", "r") as fh:
#    long_description = fh.read()
long_description = "You can find minimal working example and other information here: https://github.com/kartelj/rils-rols#readme"

if os.name=='nt':
    compile_args = ["/std:c++17", "/I rils_rols_cpp", "/O2"]
else:
    compile_args = ["-std=c++17", "-march=native", "-O3"]

ext_modules=[Pybind11Extension(
        'rils_rols_cpp', 
        sources=['rils_rols_cpp/rils_rols_cpp.cpp', 
                'rils_rols_cpp/node.cpp',
                'rils_rols_cpp/utils.cpp'], 
        extra_compile_args = compile_args
    )]

def copy_dir():
    dir_path = 'rils_rols_cpp/eigen'
    base_dir = os.path.join('.', dir_path)
    for (dirpath, _, files) in os.walk(base_dir):
        for f in files:
            print(f)
            yield os.path.join(dirpath.split('/', 1)[1], f)

setup(
    name='rils-rols',
    version='1.5.1',
    description='RILS-ROLS: Robust Symbolic Regression via Iterated Local Search and Ordinary Least Squares',
    long_description= long_description,
    long_description_content_type  = "text/markdown",
    author='Aleksandar Kartelj, Marko Đukanović, Ján Pigoš',
    author_email='aleksandar.kartelj@gmail.com',
    url='https://github.com/kartelj/rils-rols',
    packages = find_packages(),
    ext_modules=ext_modules,
    extras_require={"test": "pytest"},
    cmdclass={"build_ext": build_ext},
    zip_safe = False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License", 
        "Operating System :: OS Independent"
    ],
    python_requires= ">=3.8",
    py_modules=["rils-rols"],
    package_dir = {'rils-rols':'rils_rols'}, 
    package_data = {'' : [f for f in copy_dir()]},
    include_package_data=True,
    install_requires=["scikit-learn", "sympy", "pybind11"],
)



import math
import time
from sklearn.base import BaseEstimator
from sympy import *
from .rils_rols import RILSROLSBinaryClassifier, RILSROLSRegressor
#from joblib import Parallel, delayed
from .utils import complexity_sympy
from sklearn.model_selection import train_test_split
import warnings

warnings.filterwarnings("ignore")

class RILSROLSEnsembleBase(BaseEstimator):

    def __init__(self, classification=None, validation_size=0.5, max_fit_calls_per_estimator=100000, max_seconds_per_estimator=100, complexity_penalty=0.001, max_complexity=200, sample_size=0.1, estimator_cnt = 8, verbose=False, random_state=0):
         # true random numbers from random.org
        random_states = [5953,7138,7578,5452,1924,823,8269,5439,4188,5507,280,9610,2715,8212,4916,2026,4283,5664,9406,8120,835,6746,2970,4427,803,8092,1466,8895,9428,4106,918,2243,2575,1962,3123,4121,7806,3623,3130,2920,5796,7565,3826,758,5604,6510,8781,1040,8386,5460,6953,8030,9809,5274,6661,4712,3605,5195,9652,7768,2956,268,6707,8171,3462,8539,9377,5956,8137,9457,4624,2125,6821,4522,6634,8946,9558,9812,7225,4552,735,3539,3765,3817,5383,7864,8020,5376,1555,3367,6145,7044,8771,7486,3302,837,7596,8799,2116,4444]
        if estimator_cnt>len(random_states):
            raise Exception(f'Maximal number of estimators is {len(random_states)}')
        self.classification = classification
        self.validation_size = validation_size
        self.max_seconds = max_seconds_per_estimator
        self.max_fit_calls = max_fit_calls_per_estimator
        self.complexity_penalty = complexity_penalty
        self.max_complexity = max_complexity
        self.random_state = random_state
        self.estimator_cnt = estimator_cnt
        self.verbose = verbose
        self.sample_size = sample_size
        self.random_state = random_state
        self.base_estimators = []
        for i in range(estimator_cnt):
            if classification is True:
                est = RILSROLSBinaryClassifier(max_fit_calls=max_fit_calls_per_estimator, max_seconds=max_seconds_per_estimator,
                                                  complexity_penalty=complexity_penalty, sample_size=sample_size,verbose=verbose, random_state=random_states[i])
            elif classification is False:
                est = RILSROLSRegressor(max_fit_calls=max_fit_calls_per_estimator, max_seconds=max_seconds_per_estimator,
                                                  complexity_penalty=complexity_penalty, sample_size=sample_size,verbose=verbose, random_state=random_states[i])
            else:
                raise Exception('Estimator must be either classifier or regressor, there is no third option at the moment.')
            self.base_estimators.append(est)

    def fit(self, X, y):
        # currently, this works by selecting the best of base estimators w.r.t. performance on the validation data part
        # TODO: other options are averaging on the regression or voting on the classification
        self.start = time.time()
        X_train, X_valid, y_train, y_valid = train_test_split(X, y, test_size=self.validation_size,random_state=self.random_state)
        # now run each base regressor as a separate process 
        # TODO: this is not working because of serialization of cpp object rr_cpp probably...
        #Parallel(n_jobs=len(self.base_estimators))(delayed(reg.fit)(X_train, y_train) for reg in self.base_estimators)
        # doing only sequentially for now
        for est in self.base_estimators:
            est.fit(X_train, y_train)
        print("All base estimators have finished now")
        # checking performances on the validation set
        best_score = -math.inf
        self.best_est = None
        for est  in self.base_estimators:
            score = est.score(X_valid, y_valid)
            print(f'Score: {score} model: {est.model_string()}')
            if score>best_score:
                best_score = score
                self.best_est = est
        self.time_elapsed = time.time()-self.start
        self.model = self.best_est.model
        self.model_simp = self.best_est.model_simp
        print(f'Best simplified model is {self.model_simp} with score {best_score}')

    def check_model(self):
        if self.model is None or self.best_est is None:
            raise Exception("Cannot predict because model is not build yet. First call fit().")

    def predict(self, X):
        self.check_model()
        return self.best_est.predict(X)
    
    def score(self, X, y):
        self.check_model()
        return self.best_est.score(X, y)

    def model_string(self):
        self.check_model()
        return self.model_simp

    def fit_report_string(self):
        self.check_model()
        return "maxTime={0}\tmaxFitCalls={1}\tseed={2}\tsizePenalty={3}\tmaxComplexity={4}\tsampleShare={5}\ttotalTime={6:.1f}\tsimpSize={7}\texpr={8}\texprSimp={9}".format(
            self.max_seconds,self.max_fit_calls,self.random_state,self.complexity_penalty,self.max_complexity, self.sample_size, self.time_elapsed, complexity_sympy(self.model_simp),  self.model, self.model_simp)

class RILSROLSEnsembleBinaryClassifier(RILSROLSEnsembleBase):
    
    def __init__(self, max_fit_calls_per_estimator=100000, max_seconds_per_estimator=100, complexity_penalty=0.001, max_complexity=200, sample_size=0.1,estimator_cnt = 8, validation_size=0.5, verbose=False, random_state=0):
        super().__init__(True, validation_size=validation_size, max_fit_calls_per_estimator=max_fit_calls_per_estimator, max_seconds_per_estimator=max_seconds_per_estimator, complexity_penalty=complexity_penalty, max_complexity=max_complexity, sample_size=sample_size, estimator_cnt=estimator_cnt, verbose=verbose,  random_state=random_state)

    def predict_proba(self, X):
        self.check_model()
        return self.best_est.predict_proba(X)

class RILSROLSEnsembleRegressor(RILSROLSEnsembleBase):
    
    def __init__(self, max_fit_calls_per_estimator=100000, max_seconds_per_estimator=100, complexity_penalty=0.001, max_complexity=200, sample_size=0.1, estimator_cnt = 8, validation_size=0.5, verbose=False, random_state=0):
        super().__init__(False, validation_size=validation_size,  max_fit_calls_per_estimator=max_fit_calls_per_estimator, max_seconds_per_estimator=max_seconds_per_estimator, complexity_penalty=complexity_penalty, max_complexity=max_complexity, sample_size=sample_size, estimator_cnt=estimator_cnt, verbose=verbose,  random_state=random_state)
